###############################################################################
# Keith R. Jackson, LBNL (Based on Matt Rodriguez's initial version)
# See LBNLCopyright for copyright notice!
###############################################################################
import base64, sys, ZSI
import pyGridWare.generated.stubs.delegation_factory_service as DS
from pyGridWare.utility.scripts.client import \
    GetOptParser, GetPortKWArgs, SetUp, GetConfigParser
from ZSI.wstools.Namespaces import OASIS
from ZSI.TC import _get_global_element_declaration
from twisted.internet import reactor
from exceptions import Exception
try:
    from pyGsi import proxylib
except ImportError:
    print "Need to have pyGsi installed."
    sys.exit(-1)
# PyGsi cannot be imported if M2 isn't installed, so this is safe
from M2Crypto import X509

###class InvalidProxyException(Exception): pass
    
###############################################################################
# Constants
###############################################################################
# Represents the QName of the Resource Property that contains
# the cert chain used to delegate to the delegation service
CERTCHAIN_RP = ('http://www.globus.org/08/2004/delegationService',
                                            'CertificateChain')
#These namespaces should be moved to ZSI.wstools.Namespaces 
XMLSEC_BASE64_ENCODING_NS = \
    "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary"
XMLSEC_PKIPATH_NS = \
    "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509PKIPathv1"
DFPortType = DS.DelegationFactoryPortTypeSOAPBindingSOAP
GED = _get_global_element_declaration

###############################################################################
# Classes
###############################################################################
# XXXXXX: Does this belong in this module? Seems kind of generic (KRJ)
class QName(tuple):
    '''Use to specify a Qualified Name (namespace,name) --> prefix:name'''
    typecode = ZSI.TC.QName()

###############################################################################
# Functions
###############################################################################
def delegate_cred(lifetime=12, isFull=True, **kw):
    """Delegates a credential and Returns the EPR of the resource
       
       Contact the delegation service, constructs and signs the new
       credential, and then sends it to the delegation service. The
       service returns an EPR that can be used to reference the
       delegated credential.

       TODO: Document what keyword arguments this takes

       lifetime - Suggestion as to how long the delegated proxy should
       be valid for. Policy may override the users request. Default is
       12 hours.

       isFull - Boolean describing type of proxy to delegate. True
       is for a full proxy, and False for a limited proxy.
       Full grants all rights to the proxy, including the ability to
       generate new proxies. Limited does not allow new proxies to 
       be generated.
    """

    # -------------------------------------------
    # Check if certificate is valid.  Doesn't make any sense to
    # delegate an invalid proxy.
    # -------------------------------------------
    
    proxyfile = proxylib.get_proxy_filename()                                                         
    cert = proxylib.get_usercert(proxyfile)
    if not proxylib.check_valid(cert):
        raise proxylib.InvalidProxyException

    # -------------------------------------------

    # Get CertificatChain from DelegationFactoryService
    # 
    # The chain is stored as a ResourceProperty of the
    # delegation service.
    # -------------------------------------------

    kw['endPointReference'] = None
    dPort = DFPortType(**kw)
    getReq = DS.GetResourcePropertyRequest(QName(CERTCHAIN_RP))
    getRsp = dPort.GetResourceProperty(getReq)

    # -------------------------------------------
    # Grab BinarySecurityToken, Create X509 Stack (PKIPath)                                           
    # -------------------------------------------                                                     
    certChain = getRsp._any[0]                                                                        
    bstoken = certChain._any                                                                          
    #assert isinstance(certChain.typecode, \                                                          
    #    GED('http://www.globus.org/08/2004/delegationService',                                       
    #        'CertificateChain')), \                                                                  
    #        'Expecting CertificateChain'                                                             
    stack = X509.new_stack_from_der(base64.decodestring(bstoken))                                     
    # Grab the cert that we're going to grab the public key from
    delegation_cert = stack.pop()                                                                     
    # Create a new cert and add the public key from the cert
    # we retrieved from the resource property. The private key
    # is owned by the container we are delegating to.
    new_delegation_cert = X509.X509()                                                                 
    new_delegation_cert.set_pubkey(delegation_cert.get_pubkey())                                      
                                                                                                      
    # -------------------------------------------                                                     
    # Grab user credential, use to sign 1st cred in PKIPath.                                              
    # -------------------------------------------                                                     
    proxy_cred = proxylib.Proxy()                                                                     
    proxy_cred.read()                                                                                 
    proxy_cert = proxy_cred.getcert()                                                                 
    proxy_key = proxy_cred.getkey()                                                                   
    issuer_cert = proxy_cred.getissuer()                                                              
    # Create a stack to hold the cert chain we're going to 
    # send back to the delegation service
    bst_stack = X509.X509_Stack()                                                                     
    bst_stack.push(issuer_cert)                                                                       
    bst_stack.push(proxy_cert)                                                                        
                                                                                                      
    proxy_factory = proxylib.ProxyFactory(proxyfile, proxyfile, '', (lifetime,0))    
    proxy_factory.set_full(isFull)
    proxy_factory.set_proxycert(new_delegation_cert)   
    proxy_factory.generate() 
    new_proxy = proxy_factory.getproxy() 
    bst_stack.push(new_proxy.getcert()) 
    der_encoding = bst_stack.as_der()  
    # -------------------------------------------             
    # RequestSecurityToken                                    
    # -------------------------------------------             
                                                              
    BST = GED(OASIS.WSSE,"BinarySecurityToken").pyclass  
    bst = BST(base64.encodestring(der_encoding))  
                     
    bst._attrs = {} 
    bst._attrs['EncodingType'] = XMLSEC_BASE64_ENCODING_NS
    bst._attrs['ValueType'] = XMLSEC_PKIPATH_NS
    # Create message to send
    rstokMsg = DS.RequestSecurityTokenMsg()   
    rstokMsg._any = [bst,]  
    # Send message, should get the EPR back
    rstokMsgR = dPort.RequestSecurityToken(rstokMsg)  
    # -------------------------------------------   
    # EPR of delegated credential.   
    # -------------------------------------------  
    epr = rstokMsgR._any[0]  
    return epr  
