############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
from zope.interface import Interface, Attribute


class IWSResourceContext(Interface):
    """Represents the context of a WSRF Resource, contains 
    resource properties, URL, resourceID.
    """
    Attribute("home", "resource home (manager)")
    Attribute("url", "URL of this resource context")
    Attribute("resourceID", "resource ID of this resource context")
    
    def getAddress(self):
        """return URL of this resource context
        """
    def setAddress(self, address):
        """set URL of this resource context
        """
    def getResourceID(self):
        """return resource ID
        """
    def setResourceID(self, resourceID):
        """set resource ID
        """
    def getResourceProperties(self):
        """return Resource Properties
        """
    def setResourceProperties(self, properties):
        """set Resource Properties
        """
    def save(self):
        """invoke when resource context/properties change 
        to manually save state. 
        """


class IWSResource(Interface):
    """Represents a WSRF Resource, implies there will be
    a resource home indexed via addressing abstraction
    thru the resource home manager.
    """
    Attribute('home', 'resource home (manager)')
   
 
class IResourceHomeFactory(Interface):
    """any class providing this interface is expected to 
    instances that provide the IResourceHome interface.
    """
    
class IResourceHome(Interface):
    """Interface for ResourceHome, any class implementing this interface
    is expected to handle lifetime management of resources added.
    """
    def add(self, ctx, key=None):
        """save resource context referenced via key.
        """
    def create(self):
        """create a new key.
        """
    def delete(self, key):
        """immediately delete resource context referenced via key.
        """
    def get(self, key):
        """get resource context via key.
        """
    def setTerminationTime(self, key, terminationTime):
        """set terminationTime of resource, and schedule deletion of 
        resource context referenced via key.
        """
    def setAvatarId(self, avatarId):
        """set avatarId
        """
    def getAvatarId(self):
        """get avatarId
        """
    def getResources(self):
        """return all resource contexts.
        """
    def getResourceContext(self, ps, address):
        """return resource context.
        """

class IResourceHomeManager(IResourceHome):
    """ResourceHomeManager Interface
    """
    def newResourceHome(self):
        """return new resource home instance.
        """
    def setResourceHome(self, key, home):
        """add resource home to registry.
        """      
    def getResourceHome(self, key):
        """get resource home from registry using key.
        """
    def save(self, key=None):
        """If key is None save all resources, else save only 
        resource specified via key.
        """
    def load(self):
        """Implement this method to load up state upon initialization.
        """


class IPersistent(Interface):
    """Interface for Persistent ResourceHome, any class implementing 
    this interface is expected to save all resource.
    """
    def save(self, home, key=None):
        """If key is None save all resources, else save only 
        resource specified via key.
        """
    def load(self):
        """Implement this method to load up state upon initialization.
        """
    def setResourceManager(self, manager):
        """ResourceHomeManager instance managing this resource
        """



