###########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
# $Id:
import os
from twisted.python import log
from twisted.internet import defer, reactor
from twisted.cred.error import Unauthorized, UnauthorizedLogin
from pyGridWare.resource.interfaces import IResourceHome, IWSResource
from interfaces import IX509Credential


class X509SubjectDNChecker(object):
    
    credentialInterfaces = (IX509Credential,)
    
    def requestAvatarId(self, credentials):
        """return subject DN
        """
        subject =  credentials.get_subject()
        log.msg('X509    --  "%s"' %credentials, debug=True)
        log.msg('subject --  "%s"' %subject, debug=True)
        if not subject:
            defer.fail(UnauthorizedLogin("permission denied"))
        
        return defer.succeed(subject)
    
    
class GridMapChecker(object):
    """takes an X509Credential, uses the SubjectDN to perform
    a lookup on a grid-mapfile dict.  Using this checker limits
    an implementation to utilizing one GridMap.  You could change
    the gridmap file, but all GridMapCheckers would use the same
    one.
    
    returns the userID.
    
    class attributes:
        gridmapfile -- gridmap file
        interval -- interval for refreshing gridmap dict
        gridmap -- gridmap dictionary of subjectDN to user mapping.
    """
    credentialInterfaces = (IX509Credential,)
    gridmapfile = os.path.join(os.sep, 'etc', 'grid-security', 'grid-mapfile')
    interval = 30
    gridmap = None

    def __init__(self):
        if self.gridmap is not None:
            return

        GridMapChecker.load()
        
    @classmethod
    def load(cls, gridmapfile=None):
        """reload gridmap dict
        """
        gridmapfile = gridmapfile or cls.gridmapfile
        fd = open(gridmapfile , 'r')    
        cls.gridmap = dict(map(lambda s: s.strip('"').split('" ')[:2], fd))
        reactor.callLater(cls.interval, cls.load, gridmapfile)
    
    def requestAvatarId(self, credentials):
        """return subject DN
        """
        if self.gridmap is None:
            defer.fail(UnauthorizedLogin(
                           "permission denied: no gridmap-file"))
            
        subject =  credentials.get_subject()
        log.msg('gridmap subject "%s"', subject, debug=True)
        if not subject:
            defer.fail(UnauthorizedLogin('permission denied: "%s" no gridmap entry'))
            
        if not self.gridmap.has_key(subject):
            #form Subject DNs by removing the last CN field in the
            #subject, continue this until subject name matches or 
            #there are no more CN fields
            while True:
                subject_name_list = subject.split("/")
                subject_name_list.pop()
                subject_name_list.pop(0)
                new_subject_list = map(lambda x: "/" +x,subject_name_list)
                subject = "".join(new_subject_list)
                if self.gridmap.has_key(subject):
                    break
                try:
                    subject.index("CN")
                except ValueError:
                    defer.fail(UnauthorizedLogin(
                       'authentication failed, no gridmap entry for subjectDN "%s"' %
                       subject))
            else:
                defer.fail(UnauthorizedLogin(
                   'authentication failed, no gridmap entry for subjectDN "%s"' %
                   subject))
            
        #TODO: should probably check if user exists?
        user = self.gridmap.get(subject)
        if not user:
            defer.fail(UnauthorizedLogin(
               'authentication failed, bad gridmap entry for subjectDN "%s"' %
               subject))
            
        return defer.succeed(user)


