###########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################

# $Id: resource.py 1335 2007-03-11 20:59:28Z boverhof $
from M2Crypto import m2, X509
from M2Crypto.SSL.TwistedProtocolWrapper import TLSProtocolWrapper

# twisted 
from twisted.web.resource import Resource, IResource
from twisted.python import log, failure
from twisted.cred.portal import Portal, IRealm
from twisted.cred.error import Unauthorized, UnauthorizedLogin
from zope.interface import implements

from pyGridWare.resource.interfaces import IResourceHome,\
    IResourceHomeManager, IWSResource
    
from checkers import X509SubjectDNChecker
from credentials import X509Credential


class ResourceRealm(object):
    """Realm for a single resource/service.
    """
    implements(IRealm)

    def __init__(self, resource, callback=None):
        """ Leaf resource
        Parameters:
            resource -- protected resource
            callback -- Called with avatarId and interface, returns True/False
        """
        self.resource = resource
        self.callback = callback

    def requestAvatar(self, avatarId, mind, *interfaces):
        """
        avatarId -- subjectDN of peer certificate.
        mind --
        interfaces -- 
        """
        logout = lambda: None
        if IResource in interfaces:
            log.msg("Avatar IResource: %s" %avatarId, debug=True)
            
            if self.callback is not None and not self.callback(avatarId, *interfaces):
                raise Unauthorized("service unavailable to avatarId(%s)" %
                                   avatarId)

            return (IResource, self.resource, logout)

        if IResourceHome in interfaces:
            log.msg("Avatar(%s) IResourceHome: %s" %
                    (avatarId, self.resource), debug=True)

            if self.callback is not None and not self.callback(avatarId, *interfaces):
                raise Unauthorized("service unavailable to avatarId(%s)" %
                                   avatarId)
            
            if IWSResource.providedBy(self.resource):
                manager = self.resource.home
                home = manager.getResourceHome(avatarId)
                if home is None:
                    home = manager.newResourceHome()
                    manager.setResourceHome(avatarId, home)
    
                return (IResourceHome, home, logout)
                
            log.err("resource must provide IWSResource interface")

        raise NotImplementedError("Do not support any of the interfaces %s" %
            (interfaces,))


class ResourcePortal(Resource):
    """ portal login mechanism.  Add credential to request
    after login.
    
    resource -- avatar, or actual Resource
    """
    
    def __init__(self, resource, checkers=(X509SubjectDNChecker(),), 
                 realmFactory=ResourceRealm, callback=None):
        """
        Parameters:
            resource -- protected twisted.web.resource.Resource
            checkers -- 
            callback -- Called with avatarId and interface, returns True/False
        """
        self.portal = Portal(realmFactory(resource, callback=callback), checkers=checkers)
        self.interfaces = (IResource,)
        
        if not IResource.providedBy(resource):
            raise TypeError('resource must provide interface %s' %
                            IWSResource)
        
        if IWSResource.providedBy(resource):
            resource.home.portal = self.portal
            self.interfaces = (IResource, IWSResource)
            
        Resource.__init__(self)

    def _login(self, request):
        # TODO: Create a HTTPRequest Adapter, 
        # currently just setting credential attribute
        #
        
        #TODO: Change this to a WSRF "BaseFault", or catch it etc,
        #  so it gets serialized.
        if not hasattr(request.transport, 'ssl'):
            raise UnauthorizedLogin("must use ssl to access this resource")
        
        # TODO: Hide "portal" and "credential" from actual service
        x = X509.X509(m2.ssl_get_peer_cert(request.transport.ssl),1)
        request.credential = credential = X509Credential(x)
        portal = self.portal
        
        d = portal.login(credential, None, *self.interfaces)
        
        from twisted.python.failure import Failure
        if isinstance(d.result, Failure):
            d.result.raiseException()
        
        return d.result
        
    def getChild(self, path, request):
        """ If attempt to grab one of my children, login, and
        delegate to underlying resource.
        """
        i,avatar,logout = self._login(request)
        avatar.server = self.server
        if avatar.isLeaf:
            avatar.postpath.insert(0, request.prepath.pop())
            return avatar
        
        log.msg('AVATAR(%s):   %s' %(avatar, avatar.__dict__), debug=True)
        child = avatar.getChildWithDefault(path, request)
        log.msg('CHILD(%s):   %s' %(child, child.__dict__), debug=True)
        return child
    
    def render(self, request):
        """ If this node is rendered, login, and delegate to 
        underlying resource.
        """
        #i,avatar,logout = self._login(IX509Request(request))
        i,avatar,logout = self._login(request)
        avatar.server = self.server
        return avatar.render(request)
    

