############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import os, sys, urlparse, optparse, types
from twisted.python import log
from twisted.internet import reactor
from ConfigParser import ConfigParser, NoOptionError
from ZSI.wstools.Namespaces import WSA200403
from exceptions import Exception

SSL = SSLMETHOD = None   
GSI = False
_proxylib = None
try: 
    import pyGridWare.utility.m2crypto
    from M2Crypto import SSL
    SSLMETHOD = SSL.SSLv3_METHOD
    from pyGsi import proxylib as _proxylib
    GSI = True
except ImportError:
    pass

"""
If M2Crypto is unavailable, SSL features are disabled.  

If pyGsi is unavailable, GSI features are disabled.  This includes
use of proxy, default search locations /etc/grid-security

GSI Behavior:

    All options override default behavior.  If cert/key
aren't specified, use proxy in default location, else look for
'~/.globus/user[cert,key].pem.  If calist is unspecified use 
'/etc/grid-security/*.0"

"""

###########################################################################
# PRIVATE FUNCTIONS
###########################################################################
def _setup_secure_message(options):
    """return signature handler
    """
    from pyGridWare.security.m2.ResourceHome import ManagerHome
    from pyGridWare.plugin import GetPlugin
    from pyGridWare.resource import IResourceHomeFactory
    
    #NOTE: Initialize the ManagerHome Singleton
    home = ManagerHome.getInstance(GetPlugin(IResourceHomeFactory))

    from pyGridWare.security.m2.M2SignatureHandler import M2SignatureHandler
    handler = M2SignatureHandler()
    handler.setSecureMessage()
    return handler

def _setup_secure_conversation(options):
    """return signature handler
    """
    from pyGridWare.security.m2.ResourceHome import ManagerHome
    from pyGridWare.plugin import GetPlugin
    from pyGridWare.resource import IResourceHomeFactory
    
    #NOTE: Initialize the ManagerHome Singleton
    home = ManagerHome.getInstance(GetPlugin(IResourceHomeFactory))

    from pyGridWare.security.m2.Authenticate import SecureConversationClient
    from pyGridWare.security.m2.M2SignatureHandler import M2SignatureHandler

    ctx = _get_ssl_ctx(options)
    kwc = dict(wsAddressURI=options.wsAddressURI, 
              url=options.m2secureConv, #delegate=options.delegate,
          )

    keyname = SecureConversationClient().exchange(ctx, **kwc)
    handler = M2SignatureHandler()
    handler.setSecureConversation(keyname)
    return handler

def _get_calist_from_cp():
    cp = GetConfigParser()
    if cp.has_section('security'):
        calist = \
            filter(\
                lambda s: s, 
                    map(\
                         lambda s: s.strip(), 
                         GetConfigParser().get('security', 'ca').split(' ')
                    )
            )
        return calist

    return

def _set_authorization(options):
    from pyGridWare.utility.m2crypto import OnConnectHostAuthz,\
        OnConnectSelfAuthz, OnConnectNoAuthz
        
    if options.authz is None:
        OnConnectNoAuthz(); return

    if options.authz == 'host':
        OnConnectHostAuthz(); return

    if options.authz == 'self':
        OnConnectSelfAuthz(); return


def _get_client_context_factory(options):
    from pyGridWare.utility.m2crypto import ClientContextFactory

    if not GSI:
        cert,key,calist = options.cert,options.key,options.calist

        if not (cert and key):
            raise RuntimeError('For Proxy Support install pyGsi, else specify cert/key')

        if not os.path.isfile(cert):
            raise RuntimeError, 'Missing Certificate "%s"' %cert

        if not os.path.isfile(key):
            raise RuntimeError, 'Missing Key "%s"' %key

        if isinstance(calist, basestring):
            calist = calist.split(',')
   
        return ClientContextFactory(options.key, options.cert, 
                sslmethod=SSLMETHOD, calist=calist)

    #TODO: ?? return ClientContextFactory( _get_ssl_ctx(options) )
    #  probably best to move all this logic there or into a GSI lib.

    cert,key,calist = _get_gsi_credentials(options)
    return ClientContextFactory(key, cert, sslmethod=SSLMETHOD, calist=calist)


def _get_ssl_ctx(options):
    cert,key,calist = _get_gsi_credentials(options)
    ctx = SSL.Context('sslv3')
    ctx.set_info_callback()        
    ctx.load_cert_chain(cert, keyfile=key)
    
    for ca in calist: 
        ctx.load_verify_locations(cafile=ca)

    return ctx


def _get_gsi_credentials(options):
    """return filename of cert/key pair
    """
    cert,key,calist = options.cert, options.key, options.calist

    calist = options.calist 
    if calist:
        if isinstance(calist, basestring):
            calist = options.calist.split(',')
    else:
        calist = []

    if not cert:
        proxy,cert,key = None, _proxylib.get_proxy_filename(), None
        try:
            proxy = _proxylib.get_usercert(cert)
        except:
            # TODO: WINDOWS PROBLEMS
            cert = os.path.expanduser('~/.globus/usercert.pem')
            key = os.path.expanduser('~/.globus/userkey.pem')
        else:
            if not _proxylib.check_valid(proxy):
                raise _proxylib.InvalidProxyException

        if not os.path.isfile(cert):
            raise RuntimeError, 'GSI: no certificate "%s"' %cert

        if key and not os.path.isfile(key):
            raise RuntimeError, 'GSI: no key "%s"' %key

    # TODO: Where is it on Windows?
    capath = os.path.expanduser('~/.globus/certificates')
    if not os.path.isdir(capath):
        capath = os.path.expanduser('/etc/grid-security/certificates')
        if not os.path.isdir(capath):
            log.msg('GSI: no trusted CA certificates "%s"' %capath, debug=True)
            return cert,key,calist 
       
    calist += [os.path.join(capath,i) for i in filter(lambda f: f.endswith('.0'), os.listdir(capath))]

    return cert,key,calist


###########################################################################
# SCRIPT FUNCTIONS
###########################################################################
CP = None
def GetConfigParser(f='client-config.txt'):
    '''Return the global ConfigParser instance.
    '''
    global CP
    if CP is None:
        CP = ConfigParser()
        CP.read(f)
    return CP


def GetOptParser(test, **kw):
    '''Return an OptionParser instance with a few options set from a config file
    test -- configuration file name for test.
    '''
    scheme = 'https'
    cp = GetConfigParser()
    try:
        path = cp.get('Paths', test)
        host = cp.get('Location', 'host')
        port = cp.getint('Location', 'port')
        debug = cp.get('logging', 'debug')
        if not cp.getint('security', 'ssl'):
            scheme = 'http'
    except NoOptionError, ex:
        url = None
    else:
        url = urlparse.urlunparse((scheme, '%s:%d'%(host,port), path, '', '', ''))

    op = GetBasicOptParser(debug=debug, url=url, **kw)
    if op.has_option('--calist'):
        ca_opt = op.get_option('--calist')
        calist  = _get_calist_from_cp()
        if calist: ca_opt.default = _get_calist_from_cp()

    return op

def GetBasicOptParser(debug=1, url=None, **kw):
    '''Return an OptionParser instance with a few options.
    '''
    op = optparse.OptionParser(**kw)

    op.add_option("-u", "--url", action="store", dest="url",
        help="location of service", default=url)
    op.add_option("-d", "--debug",
        action="store", dest="debug", default=debug,
        help="Debug Messages to file, if 1 send to stdout, if 0 then off.")
    op.add_option("-w", "--wsAddressURI",
        action="store", dest="wsAddressURI", default=WSA200403.ADDRESS,
        help="Specify URI of ws-addressing version to use.")

    if SSL:
        op.add_option("-c", "--cert", action="store", dest="cert", 
            default=None, help="file path of certificate")

        op.add_option("-k", "--key", action="store", dest="key", 
            default=None, help="file path of key")

        op.add_option("-l", "--calist", action="store", dest="calist", 
            default=None, help="comma separated list of ca certificates")

        op.add_option('-a', '--authz', action='store', dest='authz', 
            default=None, choices=('self', 'host',), 
            help="choose authorization, self or host.")
    else:
        log.msg('missing M2Crypto, features requiring SSL are unavailable', debug=True)

    if GSI:
        op.add_option("-v", "--m2secureConv", action="store", 
            dest="m2secureConv", default=None,
            help="To use secure conversation, provide URL to Authorization service, and negotiate key used to sign/verify messages.")

        op.add_option("-x", "--m2secureMessage", action="store_true", 
            dest="m2secureMessage", default=False, 
            help="use secure message to sign messages.")
    else:
        log.msg('missing pyGsi, features requiring GSI are unavailable', debug=True)

    return op


def GetFileDescriptor(filename):
    '''Returns a file descriptor or None.
    filename -- str
    '''
    if type(filename) is types.FileType:
        return filename
    if not filename or filename == '0':
        return None
    elif filename == '1' or filename == 1:
        return sys.stdout
    return open(filename, 'w+')


def GetPortKWArgs(options):
    '''returns a set of key word arguments for port initialization.
    options -- Options instance.
    '''
    kw = dict(wsAddressURI=options.wsAddressURI, url=options.url)

    # message layer security
    if GSI and options.m2secureConv:
        kw['sig_handler'] = _setup_secure_conversation(options)
    elif GSI and options.m2secureMessage:
        kw['sig_handler'] = _setup_secure_message(options)

    if not options.url.startswith('https'):
        return kw

    # transport security
    _set_authorization(options)
    kw['contextFactory'] = _get_client_context_factory(options)
     
    return kw


def SetUp(options):
    '''General Script SetUp.
    options -- Options instance.
    '''
    fd = GetFileDescriptor(options.debug)
    if fd is not None:
        log.startLogging(fd)

