#!/usr/bin/env python
##########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import optparse, os, sys, signal, time, warnings, unittest
from urlparse import urlparse, urlunparse
from distutils.core import setup
from distutils.dist import Distribution
from distutils.command.clean import clean as _clean
from distutils.command.build import build as _build
from distutils.dir_util import remove_tree
from os.path import isfile
from os.path import join
from ConfigParser import ConfigParser

from twisted.internet import reactor
from twisted.python import log

from ZSI import ZSIException
from ZSI.wstools.WSDLTools import WSDLReader
from ZSI.TC import _get_global_element_declaration as GED

from pyGridWare.utility.generate.Utility import CLIENT,SERVICE,\
    SERVICE_STUBS,TYPES,PROPERTIES,RESOURCE, NameResolver as NR
from pyGridWare.utility.generate.Modules import WriteServiceModules,\
    WriteClientStubModule, WriteTypesModule

from pyGridWare.wsrf.notification import SubscribeUtility
from pyGridWare.wsrf.notification.NotificationConsumer import \
    NotificationConsumer, NotificationConsumerService
    
import pyGridWare.wsrf.faults

from counter import CounterTestSuite, CounterTestCase as _Base
    
"""
GLOBALS:
  WSDL    -- WSDL to generate types, stubs from.
  SECTION -- Ends up being the generated module representing this WSDL.
"""
WSDL = join(*'../../share/schema/core/tests/performance/counter_service.wsdl'.split('/'))
TOP = 'performance'
PATH = '/wsrf/services/PerformanceCounter'
_TARGETNAMESPACE = "http://perf.counter.com"

class CounterTestCase(_Base):
    path  = '/wsrf/services/PerformanceCounter'

    def setUp(self):
        import generated.performance.stubs.performance as STUB
        loc = STUB.CounterServiceLocator()
        kw = self.client_kw
        self.port = loc.getCounterPortType(endPointReference=self.epr, **kw)
        self.shouldStop = True


# Counter Specific
class AddTestCase(CounterTestCase):
    """Counter Add
    """
    def getMethodToCall(self):
        return self.port.add

    def NewRequestInstance(self):
        import generated.performance.stubs.performance as STUB
        return STUB.AddInputMessage(10)

    def setUp(self):
        if self.consumer is not None: self.consumer.reset()
        self.failUnless(self.epr is not None)
        CounterTestCase.setUp(self)

    def tearDown(self):
        import generated.performance.stubs.performance as STUB
        self.failUnless(isinstance(self.rsp, STUB.AddOutputMessage))

        if self.consumer is not None:
             reactor.callLater(2, CheckForNotification, self)

        CounterTestCase.tearDown(self)

    def run(self, result=None, consumer=None, **kw):
        self.consumer = consumer
        CounterTestCase.run(self, result, **kw)



class CreateTestCase(CounterTestCase):
    """WS-
    """
    def getMethodToCall(self):
        return self.port.createCounter
        
    def getEPR(self):
        """if factory method, return EPR from response.
        """
        if self.rsp is None: return
        self.failUnless(getattr(self.rsp,'_EndpointReference',None) is not None, 
            'no EPR in create response.')
        return self.rsp._EndpointReference

    def NewRequestInstance(self):
        import generated.performance.stubs.performance as STUB
        return STUB.CreateCounterRequest()
        
    def tearDown(self):
        import generated.performance.stubs.performance as STUB
        self.failUnless(isinstance(self.rsp, STUB.CreateCounterResponse))
        CounterTestCase.tearDown(self)


# WS-Resource
class GetResourcePropertyTestCase(CounterTestCase):
    def __init__(self, value, **kw):
        """Parameters:
        value -- expected value
        """
        CounterTestCase.__init__(self, **kw)
        self.__value = value
        
    def getMethodToCall(self):
        return self.port.GetResourceProperty

    def NewRequestInstance(self):
        import generated.performance.stubs.performance as STUB
        return STUB.GetResourcePropertyRequest((_TARGETNAMESPACE,"Value"))
        
    def tearDown(self):
        import generated.performance.stubs.performance as STUB
        self.failUnless(isinstance(self.rsp, STUB.GetResourcePropertyResponse))
        #print "RESPONSE: ", self.rsp
        #self.failUnless(self.rsp.Any[0] == self.__value)
        CounterTestCase.tearDown(self)


class SetResourcePropertiesTestCase(CounterTestCase):
    def __init__(self, value, **kw):
        """Parameters:
        value -- set RP to this value
        """
        CounterTestCase.__init__(self, **kw)
        self.__value = value

    def getMethodToCall(self):
        return self.port.SetResourceProperties

    def NewRequestInstance(self):
        import generated.performance.stubs.performance as STUB
        pyobj = STUB.SetResourcePropertiesRequest()
        klass = GED(_TARGETNAMESPACE,"Value").pyclass

        up = pyobj.new_Update()
        pyobj.Update = [up]
        up.Any = [klass(self.__value)]
        return pyobj
        
    def tearDown(self):
        import generated.performance.stubs.performance as STUB
        self.failUnless(isinstance(self.rsp, STUB.SetResourcePropertiesResponse))
        CounterTestCase.tearDown(self)


def PerformanceTestSuite():
    return CounterTestSuite(tests=(CreateTestCase(), AddTestCase(),
                                   SetResourcePropertiesTestCase(1918),
                                    GetResourcePropertyTestCase(1918)))
