#!/usr/bin/env python
##########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import optparse, os, sys, signal, time, warnings, unittest
from urlparse import urlparse, urlunparse
from distutils.core import setup
from distutils.dist import Distribution
from distutils.command.clean import clean as _clean
from distutils.command.build import build as _build
from distutils.dir_util import remove_tree
from os.path import isfile
from os.path import join
from ConfigParser import ConfigParser

from twisted.internet import reactor
from twisted.python import log

from ZSI import ZSIException
from ZSI.wstools.WSDLTools import WSDLReader

from pyGridWare.utility.generate.Utility import CLIENT,SERVICE,\
    SERVICE_STUBS,TYPES,PROPERTIES,RESOURCE, NameResolver as NR
from pyGridWare.utility.generate.Modules import WriteServiceModules,\
    WriteClientStubModule, WriteTypesModule

from pyGridWare.wsrf.notification import SubscribeUtility
from pyGridWare.wsrf.notification.NotificationConsumer import \
    NotificationConsumer, NotificationConsumerService
    
import pyGridWare.wsrf.faults
from suites import WSRFTestSuite, WSRFTestCase, SubscribeMarker

from pyGridWare.addressing.AddressingUtils import EndpointReference as EPR
    
"""
GLOBALS:
  WSDL    -- WSDL to generate types, stubs from.
"""
WSDL = join("..", "..", "share", "schema", "core", "samples", "counter", 
                                                    "counter_service.wsdl")


def Initialize(root):
    """Special function, called by twistd wsrf command to set up the 
    root and allow us to configure various aspects of the resources
    """
    from pyGridWare.resource import IResourceHomeFactory, IPersistent
    from pyGridWare.plugin import SetDefaultPlugin, GetPlugins, GetPlugin, DEFAULT
    from pyGridWare.plugins.ServiceMaker import setUpDropIns

    for p in GetPlugins(IResourceHomeFactory):
        if not IPersistent.providedBy(p): continue
        SetDefaultPlugin(IResourceHomeFactory, p)
        break

    setUpDropIns(root)


def CheckForNotification(test):
    test.failUnless(test.consumer.called(), 'Waited 2seconds For Notification')


# Base class for all counter test cases
class CounterTestCase(WSRFTestCase):
    """
    class attributes:
        path -- the path portion of the url of the service
    """
    path = '/wsrf/services/CounterService'

    def setUp(self):
        import generated.counter_service.stubs.counter_service as client
        loc = client.CounterServiceLocator()
        kw = self.client_kw
        self.port = loc.getCounterPortTypePort(endPointReference=self.epr, **kw)
        self.shouldStop = True
        
    def tearDown(self):
        self.shouldStop = False
        


# Counter Specific
class AddTestCase(CounterTestCase):
    """Counter Add
    """
    def getMethodToCall(self):
        return self.port.add

    def NewRequestInstance(self):
        import generated.counter_service.stubs.counter_service
        return generated.counter_service.stubs.counter_service.AddInputMessage(10)

    def setUp(self):
        if self.consumer is not None: self.consumer.reset()
        self.failUnless(self.epr is not None)
        CounterTestCase.setUp(self)
        
    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, 
            generated.counter_service.stubs.counter_service.AddOutputMessage))

        if self.consumer is not None:
             reactor.callLater(2, CheckForNotification, self)

        CounterTestCase.tearDown(self)

    def run(self, result=None, consumer=None, **kw):
        self.consumer = consumer
        CounterTestCase.run(self, result, **kw)


class SavedEPR(CounterTestCase):

    def runTest(self):
        epr = EPR.load(open('persistent_epr.xml'))
        self.setEPR(epr)
        return


class CreateTestCase(CounterTestCase):
    """WS-
    """
    def getMethodToCall(self):
        return self.port.createCounter
        
    def getEPR(self):
        """if factory method, return EPR from response.
        """
        if self.rsp is None: return
        self.failUnless(getattr(self.rsp,'_EndpointReference',None) is not None, 
            'no EPR in create response.')
        return self.rsp._EndpointReference

    def NewRequestInstance(self):
        import generated.counter_service.stubs.counter_service
        return generated.counter_service.stubs.counter_service.CreateCounterRequest()
        
    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.CreateCounterResponse))
        epr = self.getEPR()
        EPR.save(open('persistent_epr.xml', 'w'), epr)
        CounterTestCase.tearDown(self)


# -----------------------------------------------
# WS-ResourceProperties
# -----------------------------------------------
class GetResourcePropertyTestCase(CounterTestCase):
    def getMethodToCall(self):
        return self.port.GetResourceProperty

    def NewRequestInstance(self):
        import generated.counter_service.stubs.counter_service
        return generated.counter_service.stubs.counter_service.GetResourcePropertyRequest(("http://counter.com","Value"))
        
    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.GetResourcePropertyResponse))
        CounterTestCase.tearDown(self)


class GetMultipleResourcePropertiesTestCase(CounterTestCase):
    def getMethodToCall(self):
        return self.port.GetMultipleResourceProperties

    def NewRequestInstance(self):
        import generated.counter_service.stubs.counter_service
        msg = generated.counter_service.stubs.counter_service.GetMultipleResourcePropertiesRequest()
        msg.ResourceProperty = [("http://counter.com","Value"), 
            ("http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceLifetime-1.2-draft-01.xsd","CurrentTime"),]
        return msg

    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.GetMultipleResourcePropertiesResponse))
        CounterTestCase.tearDown(self)


class QueryResourcePropertiesTestCase(CounterTestCase):
    def getMethodToCall(self):
        return self.port.QueryResourceProperties

    def NewRequestInstance(self):
        from ZSI.wstools.Namespaces import DSIG
        import generated.counter_service.stubs.counter_service
        msg = generated.counter_service.stubs.counter_service.QueryResourcePropertiesRequest()
        msg._QueryExpression = filter(lambda what: what.aname=='_QueryExpression', msg.typecode.ofwhat)[0].pyclass()
        msg._QueryExpression._text  = "contains(/*/*[namespace-uri()='http://counter.com' and local-name()='operations supported'],\"Value\")"
        msg._QueryExpression._attrs = {\
            'Dialect':DSIG.XPATH,
        }
        return msg

    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.QueryResourcePropertiesResponse))
        self.failUnless(self.rsp)
        CounterTestCase.tearDown(self)

    def runTest(self):
        import generated.counter_service.stubs.counter_service
        CounterTestCase.runTest(self)
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.QueryResourcePropertiesResponse))
        self.failUnless(self.rsp._text == 'true')


# -----------------------------------------------
# WS-Lifetime
# -----------------------------------------------
class DestroyTestCase(CounterTestCase):
    """WS-
    """
    def getMethodToCall(self):
        return self.port.Destroy

    def NewRequestInstance(self):
        import generated.counter_service.stubs.counter_service
        return generated.counter_service.stubs.counter_service.DestroyRequest()
        
    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.DestroyResponse))
        CounterTestCase.tearDown(self)
        self.shouldStop = True


class SetTerminationTimeTestCase(CounterTestCase):
    """after this returns resource should be destroyed.
    """
    def getMethodToCall(self):
        return self.port.SetTerminationTime

    def NewRequestInstance(self):
        import generated.counter_service.stubs.counter_service
        msg = generated.counter_service.stubs.counter_service.SetTerminationTimeRequest()
        msg._RequestedTerminationTime = time.gmtime(time.time()+2)
        return msg
        
    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.SetTerminationTimeResponse))
        CounterTestCase.tearDown(self)
        self.shouldStop = False
        
    def runTest(self):
        """After returns resource should be destroyed.
        """
        CounterTestCase.runTest(self)
        time.sleep(10)
        
        
# -----------------------------------------------
# WS-BaseNotification
# -----------------------------------------------
class SubscribeTestCase(CounterTestCase, SubscribeMarker):
    """WS-BaseNotification
    """
    path = '/wsrf/services/CounterService'
    class NCService(NotificationConsumerService):
        def __init__(self, *args, **kw):
            self.reset()
            NotificationConsumerService.__init__(self, *args, **kw)
        def wsa_Notify(self, ps, address, **kw):
            self._called = True
            response = NotificationConsumerService.wsa_Notify(self, ps, address)
            return response,None

        def reset(self): self._called = False
        def called(self): return self._called

    def __init__(self, methodName='runTest', epr=None):
        CounterTestCase.__init__(self, methodName=methodName)
        self.consumer = NotificationConsumer(notificationClass=SubscribeTestCase.NCService)

    def getMethodToCall(self):
        return self.port.Subscribe

    def NewRequestInstance(self):
        import generated.counter_service.stubs.counter_service
        msg = generated.counter_service.stubs.counter_service.SubscribeRequest()
        SubscribeUtility.SimpleTopicExpression(msg, self.consumer, 
            namespace="http://counter.com", name="Value")
        return msg

    def setUp(self):
        self.consumer.start()
        CounterTestCase.setUp(self)

    def tearDown(self):
        import generated.counter_service.stubs.counter_service
        self.failUnless(isinstance(self.rsp, generated.counter_service.stubs.counter_service.SubscribeResponse))
        CounterTestCase.tearDown(self)
        self.shouldStop = False


class ResourceUnknownTestCase(CounterTestCase):
    
    def __init__(self, klass, **kw):
        unittest.TestCase.__init__(self, **kw)
        self.wrappedTest = klass(**kw)

    def getEPR(self):
        return self.wrappedTest.getEPR()

    def setEPR(self, epr):
        self.wrappedTest.setEPR(epr)

    def NewRequestInstance(self):
        raise NotImplementedError, 'this method is not implemented'
        
    def setUp(self):
        self.wrappedTest.client_kw = self.client_kw
        self.wrappedTest.setUp()
        
    def tearDown(self):
        self.shouldStop = True
        
    def runTest(self):
        self.failUnlessRaises(ZSIException, self.wrappedTest.runTest)


def start():
    return \
        WSRFTestSuite(\
            tests=(CreateTestCase(), AddTestCase(),),
        )

def end():
    return \
        WSRFTestSuite(\
            tests=(SavedEPR(), AddTestCase(),  DestroyTestCase()),
        )



CounterTestSuite = WSRFTestSuite
